"""
********************************************************************************
compas_xr.project
********************************************************************************

This package contains classes to manage the XR project including the assembly
and building plan.

.. currentmodule:: compas_xr.project

Classes
-------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    ProjectManager
    AssemblyExtensions
    BuildingPlanExtensions

"""

from compas_xr.project.assembly_extensions import AssemblyExtensions
from compas_xr.project.buildingplan_extensions import BuildingPlanExtensions
from compas_xr.project.project_manager import ProjectManager

__all__ = ["ProjectManager", "AssemblyExtensions", "BuildingPlanExtensions"]



"""
********************************************************************************
compas_xr.ghpython
********************************************************************************

This package contains classes to ease the usage of COMPAS XR from within Grasshopper.

.. currentmodule:: compas_xr.ghpython

Classes
-------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    AppSettings
    FirebaseConfig
    MqttMessageOptionsXR
    TrajectoryResultManager

"""

from .app_settings import AppSettings
from .firebase_config import FirebaseConfig
from .options import MqttMessageOptionsXR
from .trajectory_manager import TrajectoryResultManager

__all__ = ["AppSettings", "FirebaseConfig", "MqttMessageOptionsXR", "TrajectoryResultManager"]



import json
import os
from copy import deepcopy

from compas.data import json_dump


class StorageInterface(object):
    """
    The StorageInterface class serves as the shared interface for Storage classes that
    operate in IronPython and Python 3.0.

    Methods within this class are designed to rely on shared interfaces that are implemented in child classes.
    """

    def construct_reference(self, cloud_file_name):
        raise NotImplementedError("Implemented on child classes")

    def construct_reference_with_folder(self, cloud_folder_name, cloud_file_name):
        raise NotImplementedError("Implemented on child classes")

    def construct_reference_from_list(self, cloud_path_list):
        raise NotImplementedError("Implemented on child classes")

    def upload_data_to_reference(self, data, storage_reference, pretty=True):
        raise NotImplementedError("Implemented on child classes")

    def get_data_from_reference(self, storage_refrence):
        raise NotImplementedError("Implemented on child classes")

    def upload_bytes_to_reference_from_local_file(self, file_path, storage_reference):
        raise NotImplementedError("Implemented on child classes")

    def upload_data(self, data, cloud_file_name, pretty=True):
        """
        Uploads data to the Firebase Storage under specified cloud file name.

        Parameters
        ----------
        data : Any
            The data to be uploaded, needs to be JSON serializable.
        cloud_file_name : str
            The name of the reference under which the data will be stored file type should be specified.(ex: .json)
        pretty : bool, optional
            A boolean that determines if the data should be formatted for readability. Default is True.

        Returns
        -------
        None

        """
        storage_reference = self.construct_reference(cloud_file_name)
        self.upload_data_to_reference(data, storage_reference, pretty)

    def upload_data_from_json(self, path_local, pretty=True):
        """
        Uploads data to the Firebase Storage from JSON file.

        Parameters
        ----------
        path_local : str (path)
            The local path at which the JSON file is stored.
        pretty : bool, optional
            A boolean that determines if the data should be formatted for readability. Default is True.

        Returns
        -------
        None

        """
        if not os.path.exists(path_local):
            raise Exception("path does not exist {}".format(path_local))
        with open(path_local) as file:
            data = json.load(file)
        cloud_file_name = os.path.basename(path_local)
        storage_reference = self.construct_reference(cloud_file_name)
        self.upload_data_to_reference(data, storage_reference, pretty)

    def upload_data_to_folder(self, data, cloud_folder_name, cloud_file_name, pretty=True):
        """
        Uploads data to the Firebase Storage under specified cloud folder name in cloud file name.

        Parameters
        ----------
        data : Any
            The data to be uploaded, needs to be JSON serializable.
        cloud_folder_name : str
            The name of the folder under which the data will be stored.
        cloud_file_name : str
            The name of the reference under which the data will be stored file type should be specified.(ex: .json)
        pretty : bool, optional
            A boolean that determines if the data should be formatted for readability. Default is True.

        Returns
        -------
        None

        """
        storage_reference = self.construct_reference_with_folder(cloud_folder_name, cloud_file_name)
        self.upload_data_to_reference(data, storage_reference, pretty)

    def upload_data_to_deep_reference(self, data, cloud_path_list, pretty=True):
        """
        Uploads data to the Firebase Storage under specified reference names in list order.

        Parameters
        ----------
        data : Any
            The data to be uploaded, needs to be JSON serializable.
        cloud_path_list : list of str
            The list of reference names under which the data will be stored file type should be specified.(ex: .json)
        pretty : bool, optional
            A boolean that determines if the data should be formatted for readability. Default is True.

        Returns
        -------
        None

        """
        storage_reference = self.construct_reference_from_list(cloud_path_list)
        self.upload_data_to_reference(data, storage_reference, pretty)

    def upload_file_as_bytes(self, file_path):
        """
        Uploads a file as bytes to the Firebase Storage.

        Parameters
        ----------
        file_path : str
            The local path of the file to be uploaded.

        Returns
        -------
        None

        """
        if not os.path.exists(file_path):
            raise FileNotFoundError("File not found: {}".format(file_path))
        file_name = os.path.basename(file_path)
        storage_reference = self.construct_reference(file_name)
        self.upload_bytes_to_reference_from_local_file(file_path, storage_reference)

    def upload_file_as_bytes_to_deep_reference(self, file_path, cloud_path_list):
        """
        Uploads a file as bytes to the Firebase Storage to specified cloud path.

        Parameters
        ----------
        file_path : str
            The local path of the file to be uploaded.
        cloud_path_list : list of str
            The list of reference names under which the file will be stored.

        Returns
        -------
        None

        """
        if not os.path.exists(file_path):
            raise FileNotFoundError("File not found: {}".format(file_path))
        file_name = os.path.basename(file_path)
        new_path_list = deepcopy(cloud_path_list)
        new_path_list.append(file_name)

        storage_reference = self.construct_reference_from_list(new_path_list)
        self.upload_bytes_to_reference_from_local_file(file_path, storage_reference)

    def upload_files_as_bytes_from_directory_to_deep_reference(self, directory_path, cloud_path_list):
        """
        Uploads all files in specified directory as bytes to the Firebase Storage at specified cloud path in list order.

        Parameters
        ----------
        directory_path : str
            The local path of the directory in which files are stored.
        cloud_path_list : list of str
            The list of reference names under which the file will be stored.

        Returns
        -------
        None

        """
        if not os.path.exists(directory_path) or not os.path.isdir(directory_path):
            raise FileNotFoundError("Directory not found: {}".format(directory_path))
        for file_name in os.listdir(directory_path):
            file_path = os.path.join(directory_path, file_name)
            self.upload_file_as_bytes_to_deep_reference(file_path, cloud_path_list)

    # TODO: This works as it should, but I have a lot of problems with json_loads
    def get_data(self, cloud_file_name):
        """
        Retrieves data from the Firebase Storage for specified cloud file name.

        Parameters
        ----------
        cloud_file_name : str
            The name of the cloud file.

        Returns
        -------
        data : dict or Compas Class Object
            The retrieved data in dictionary format or as Compas Class Object.

        """
        storage_reference = self.construct_reference(cloud_file_name)
        return self.get_data_from_reference(storage_reference)

    # TODO: This is not working... for some reason the GetDownloadUrlAsync always results Faulted
    def get_data_from_folder(self, cloud_folder_name, cloud_file_name):
        """
        Retrieves data from the Firebase Storage for specified cloud folder name and cloud file name.

        Parameters
        ----------
        cloud_folder_name : str
            The name of the cloud folder.
        cloud_file_name : str
            The name of the cloud file.

        Returns
        -------
        data : dict or Compas Class Object
            The retrieved data in dictionary format or as Compas Class Object.

        """
        storage_reference = self.construct_reference_with_folder(cloud_folder_name, cloud_file_name)
        return self.get_data_from_reference(storage_reference)

    # TODO: This is not working... for some reason the GetDownloadUrlAsync always results Faulted
    def get_data_from_deep_reference(self, cloud_path_list):
        """
        Retrieves data from the Firebase Storage for specified cloud folder name and cloud file name.

        Parameters
        ----------
        cloud_folder_name : str
            The name of the cloud folder.
        cloud_file_name : str
            The name of the cloud file.

        Returns
        -------
        data : dict or Compas Class Object
            The retrieved data in dictionary format or as Compas Class Object.

        """
        storage_reference = self.construct_reference_from_list(cloud_path_list)
        return self.get_data_from_reference(storage_reference)

    # TODO: This worked with Frame.__data__ but not with TimberAssembly.__data__
    def download_data_to_json(self, cloud_file_name, path_local, pretty=True):
        """
        Downloads data from the Firebase Storage for specified cloud file name.

        Parameters
        ----------
        cloud_file_name : str
            The name of the cloud file.
        path_local : str (path)
            The local path at which the JSON file will be stored.
        pretty : bool, optional
            A boolean that determines if the data should be formatted for readability. Default is True.

        Returns
        -------
        None

        """
        data = self.get_data(cloud_file_name)
        directory_name = os.path.dirname(path_local)
        if not os.path.exists(directory_name):
            raise FileNotFoundError("Directory {} does not exist!".format(directory_name))
        json_dump(data=data, fp=path_local, pretty=pretty)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import os

import pyrebase
from compas.data import json_dumps

from compas_xr.realtime_database.realtime_database_interface import RealtimeDatabaseInterface


class RealtimeDatabase(RealtimeDatabaseInterface):
    """
    A RealtimeDatabase is defined by a Firebase configuration path and a shared database reference.

    The RealtimeDatabase class is responsible for initializing and managing the connection to a Firebase Realtime Database.
    It ensures that the database connection is established only once and shared across all instances of the class.

    Parameters
    ----------
    config_path : str
        The path to the Firebase configuration JSON file.

    Attributes
    ----------
    config_path : str
        The path to the Firebase configuration JSON file.
    database : Database
        The Database instance representing the connection to the Firebase Realtime Database.
    _shared_database : Database, class attribute
        The shared Database instance representing the connection to the Firebase Realtime Database.
    """

    _shared_database = None

    def __init__(self, config_path):
        self.config_path = config_path
        self._ensure_database()

    def _ensure_database(self):
        """
        Ensures that the database connection is established.
        If the connection is not yet established, it initializes it.
        If the connection is already established, it returns the existing connection.
        """
        if not RealtimeDatabase._shared_database:
            path = self.config_path

            if not os.path.exists(path):
                raise Exception("Could not find config file at path {}!".format(path))
            with open(path) as config_file:
                config = json.load(config_file)
            # TODO: Database Authorization (Works only with public databases)
            firebase = pyrebase.initialize_app(config)
            RealtimeDatabase._shared_database = firebase.database()

        if not RealtimeDatabase._shared_database:
            raise Exception("Could not initialize database!")

    def construct_reference(self, parentname):
        """
        Constructs a database reference under the specified parent name.

        Parameters
        ----------
        parentname : str
            The name of the parent under which the reference will be constructed.

        Returns
        -------
        :class: 'pyrebase.pyrebase.Database'
            The constructed database reference.

        """
        return RealtimeDatabase._shared_database.child(parentname)

    def construct_child_refrence(self, parentname, childname):
        """
        Constructs a database reference under the specified parent name & child name.

        Parameters
        ----------
        parentname : str
            The name of the parent under which the reference will be constructed.
        childname : str
            The name of the child under which the reference will be constructed.

        Returns
        -------
        :class: 'pyrebase.pyrebase.Database'
            The constructed database reference.

        """
        return RealtimeDatabase._shared_database.child(parentname).child(childname)

    def construct_grandchild_refrence(self, parentname, childname, grandchildname):
        """
        Constructs a database reference under the specified parent name, child name, & grandchild name.

        Parameters
        ----------
        parentname : str
            The name of the parent under which the reference will be constructed.
        childname : str
            The name of the child under which the reference will be constructed.
        grandchildname : str
            The name of the grandchild under which the reference will be constructed.

        Returns
        -------
        :class: 'pyrebase.pyrebase.Database'
            The constructed database reference.

        """
        return RealtimeDatabase._shared_database.child(parentname).child(childname).child(grandchildname)

    def construct_reference_from_list(self, reference_list):
        """
        Constructs a database reference under the specified refrences in list order.

        Parameters
        ----------
        reference_list : list of str
            The name of the parent under which the reference will be constructed.

        Returns
        -------
        :class: 'pyrebase.pyrebase.Database'
            The constructed database reference.

        """
        reference = RealtimeDatabase._shared_database
        for ref in reference_list:
            reference = reference.child(ref)
        return reference

    def delete_data_from_reference(self, database_reference):
        """
        Method for deleting data from a constructed database reference.

        Parameters
        ----------
        database_reference: 'pyrebase.pyrebase.Database'
            Reference to the database location where the data will be deleted from.

        Returns
        -------
        None
        """
        self._ensure_database()
        database_reference.remove()

    def get_data_from_reference(self, database_reference):
        """
        Method for retrieving data from a constructed database reference.

        Parameters
        ----------
        database_reference: 'pyrebase.pyrebase.Database'
            Reference to the database location where the data will be retreived from.

        Returns
        -------
        dict
            The retrieved data as a dictionary.

        """
        self._ensure_database()
        database_directory = database_reference.get()
        data = database_directory.val()
        data_dict = dict(data)
        return data_dict

    def stream_data_from_reference(self, callback, database_reference):
        raise NotImplementedError("Function Under Developement")

    def upload_data_to_reference(self, data, database_reference):
        """
        Method for uploading data to a constructed database reference.

        Parameters
        ----------
        data : Any
            The data to be uploaded. Data should be JSON serializable.
        database_reference: 'pyrebase.pyrebase.Database'
            Reference to the database location where the data will be uploaded.

        Returns
        -------
        None
        """
        self._ensure_database()
        # TODO: Check if this is stupid... it provides the functionality of making it work with compas objects and consistency across both child classes
        json_string = json_dumps(data)
        database_reference.set(json.loads(json_string))



"""
Component to define COMPAS XR options.

COMPAS XR v1.0.0
"""

from ghpythonlib.componentbase import executingcomponent as component

from compas_xr.ghpython import MqttMessageOptionsXR


class XrOptionsComponent(component):
    def RunScript(self, host, project_name, robot_name):
        return MqttMessageOptionsXR(host, project_name, robot_name)



"""
Application Settings.

COMPAS XR v1.0.0
"""

from ghpythonlib.componentbase import executingcomponent as component

from compas_xr.ghpython.app_settings import AppSettings
from compas_xr.project import ProjectManager


class ApplicationSettingsComponent(component):
    def RunScript(self, config_filepath, project_name, storage_folder, z_to_y_remap, write):
        if not (config_filepath):
            self.Message = "Missing Config"

        elif not (project_name):
            self.Message = "Missing Settings Data"

        else:
            app_settings = AppSettings(project_name, storage_folder, z_to_y_remap)
            pm = ProjectManager(config_filepath)
            self.Message = None

            if write:
                pm.application_settings_writer(app_settings.project_name, app_settings.storage_folder, app_settings.z_to_y_remap)



import os

from compas.geometry import Frame
from compas_timber.assembly import TimberAssembly
from compas_timber.planning import BuildingPlan
from compas_timber.planning import Step

from compas_xr.project.assembly_extensions import AssemblyExtensions
from compas_xr.realtime_database import RealtimeDatabase
from compas_xr.storage import Storage


class ProjectManager(object):
    """
    The ProjectManager class is responsible for managing project specific data and operations that involve
    Firebase Storage and Realtime Database configuration.

    Parameters
    ----------
    config_path : str
        The path to the configuration file for the project.

    Attributes
    ----------
    storage : Storage
        The storage instance for the project.
    database : RealtimeDatabase
        The realtime database instance for the project.
    """

    def __init__(self, config_path):
        if not os.path.exists(config_path):
            raise Exception("Could not create Storage or Database with path {}!".format(config_path))
        self.storage = Storage(config_path)
        self.database = RealtimeDatabase(config_path)

    def application_settings_writer(self, project_name, storage_folder="None", z_to_y_remap=False):
        """
        Uploads required application settings to the Firebase RealtimeDatabase.

        Parameters
        ----------
        project_name : str
            The name of the project where the app will look for information.
        storage_folder : str, optional
            The name of the storage folder, by default "None"
        z_to_y_remap : bool, optional
            The orientation of the object, if the obj was exported with z to y remap, by default False

        Returns
        -------
        None

        """
        data = {"project_name": project_name, "storage_folder": storage_folder, "z_to_y_remap": z_to_y_remap}
        self.database.upload_data(data, "ApplicationSettings")

    def create_project_data_from_compas(self, assembly, building_plan, qr_frames_list):
        """
        Formats data structure from COMPAS Class Objects.

        Parameters
        ----------
        assembly : :class:`compas.datastructures.Assembly` or :class:`compas_timber.assembly.TimberAssembly`
            The assembly in which data will be extracted from.
        building_plan : :class:`compas_timber.planning.BuildingPlan`
            The BuildingPlan in which data will be extracted from.
        qr_frames_list : list of :class:`compas.geometry.Frame`
            List of frames at specific locations for application localization data.

        Returns
        -------
        None

        """
        qr_assembly = AssemblyExtensions().create_qr_assembly(qr_frames_list)
        if isinstance(assembly, TimberAssembly):
            data = {
                "QRFrames": qr_assembly.__data__,
                "assembly": assembly.__data__,
                "beams": {beam.key: beam for beam in assembly.beams},
                "joints": {joint.key: joint for joint in assembly.joints},
                "building_plan": building_plan,
            }
        else:
            data = {
                "QRFrames": qr_assembly.__data__,
                "assembly": assembly.__data__,
                "parts": {part.key: part for part in assembly.parts()},
                "building_plan": building_plan,
            }
        return data

    def upload_data_to_project(self, data, project_name, data_name):
        """
        Uploads data to the Firebase RealtimeDatabase under the specified project name.

        Parameters
        ----------
        data : Any should be json serializable
            The data to be uploaded.
        project_name : str
            The name of the project under which the data will be stored.
        data_name : str
            The name of the child in which data will be stored.

        Returns
        -------
        None

        """
        self.database.upload_data_to_reference_as_child(data, project_name, data_name)

    def upload_project_data_from_compas(self, project_name, assembly, building_plan, qr_frames_list):
        """
        Formats data structure from COMPAS Class Objects and uploads them to the RealtimeDatabase under project name.

        Parameters
        ----------
        assembly : :class:`compas.datastructures.Assembly` or :class:`compas_timber.assembly.TimberAssembly`
            The assembly in which data will be extracted from.
        building_plan : :class:`compas_timber.planning.BuildingPlan`
            The BuildingPlan in which data will be extracted from.
        qr_frames_list : list of :class:`compas.geometry.Frame`
            List of frames at specific locations for application localization data.
        project_name : str
            The name of the project under which the data will be stored.

        Returns
        -------
        None

        """
        data = self.create_project_data_from_compas(assembly, building_plan, qr_frames_list)
        self.database.upload_data(data, project_name)

    def upload_qr_frames_to_project(self, project_name, qr_frames_list):
        """
        Uploads QR Frames to the Firebase RealtimeDatabase under the specified project name.

        Parameters
        ----------
        qr_frames_list : list of :class:`compas.geometry.Frame`
            List of frames at specific locations for application localization data.
        project_name : str
            The name of the project under which the data will be stored.

        Returns
        -------
        None

        """
        qr_assembly = AssemblyExtensions().create_qr_assembly(qr_frames_list)
        data = qr_assembly.__data__
        self.database.upload_data_to_reference_as_child(data, project_name, "QRFrames")

    def upload_obj_to_storage(self, path_local, storage_folder_name):
        """
        Upload an .obj file to the Firebase Storage under the specified storage folder name.

        Parameters
        ----------
        file_path : str
            The path at which the obj file is stored.
        storage_folder_name : str
            The name of the storage folder where the .obj file will be uploaded.

        Returns
        -------
        None

        """
        storage_folder_list = ["obj_storage", storage_folder_name]
        self.storage.upload_file_as_bytes_to_deep_reference(path_local, storage_folder_list)

    def upload_objs_from_directory_to_storage(self, local_directory, storage_folder_name):
        """
        Uploads all .obj files from a directory to the Firebase Storage under the specified storage folder name.

        Parameters
        ----------
        directory_path : str
            The path to the directory where the projects .obj files are stored.
        storage_folder_name : str
            The name of the storage folder where the .obj files will be uploaded.

        Returns
        -------
        None

        """
        storage_folder_list = ["obj_storage", storage_folder_name]
        self.storage.upload_files_as_bytes_from_directory_to_deep_reference(local_directory, storage_folder_list)

    def get_project_data(self, project_name):
        """
        Retrieves data from the Firebase RealtimeDatabase under the specified project name.

        Parameters
        ----------
        project_name : str
            The name of the project under which the data will be stored.

        Returns
        -------
        data : dict
            The data retrieved from the database at the point of fetching.

        """
        return self.database.get_data(project_name)

    def upload_compas_object_to_storage(self, compas_object, cloud_file_name, pretty=True):
        """
        Uploads an assembly to the Firebase Storage.

        Parameters
        ----------
        compas_object : Any
            Any compas class instance that is serializable.
        cloud_file_name : str
            The name of the cloud file. Saved in JSON format, and needs to have a .json extension.

        Returns
        -------
        None

        """
        self.storage.upload_data(compas_object, cloud_file_name, pretty=pretty)

    def get_assembly_from_storage(self, cloud_file_name):
        """
        Retrieves an assembly from the Firebase Storage.

        Parameters
        ----------
        cloud_file_name : str
            The name of the cloud file.

        Returns
        -------
        assembly : :class:`compas.datastructures.Assembly` or :class:`compas_timber.assembly.TimberAssembly`
            The assembly retrieved from the storage.

        """
        return self.storage.get_data(cloud_file_name)

    def edit_step_on_database(self, project_name, key, actor, is_built, is_planned, priority):
        """
        Edits a building plan step in the Firebase RealtimeDatabase under the specified project name.

        Parameters
        ----------
        project_name : str
            The name of the project under which the data will be stored.
        key : str
            The key of the building plan step to be edited.
        actor : str
            The actor who will be performing the step.
        is_built : bool
            A boolean that determines if the step is built.
        is_planned : bool
            A boolean that determines if the step is planned.
        priority : int
            The priority of the step.

        Returns
        -------
        None

        """
        database_reference_list = [project_name, "building_plan", "data", "steps", key, "data"]
        current_data = self.database.get_data_from_deep_reference(database_reference_list)
        current_data["actor"] = actor
        current_data["is_built"] = is_built
        current_data["is_planned"] = is_planned
        current_data["priority"] = priority
        self.database.upload_data_to_deep_reference(current_data, database_reference_list)

    def visualize_project_state_timbers(self, timber_assembly, project_name):
        """
        Retrieves and visualizes data from the Firebase RealtimeDatabase under the specified project name.

        Parameters
        ----------
        timber_assembly : :class:`compas_timbers.assembly.TimberAssembly`
            The assembly in which the project is based off of: Used for part visulization.
        project_name : str
            The name of the project under which the data will be stored.

        Returns
        -------
        last_built_index : int
            The index of the last built part in the project.
        step_locations : list of :class:`compas.geometry.Frame`
            The locations of the building plan steps.
        built_human : list of :class:`compas_timber.beam.Blank`
            The parts that have been built by a human.
        unbuilt_human : list of :class:`compas_timber.beam.Blank`
            The parts that have not been built by a human.
        built_robot : list of :class:`compas_timber.beam.Blank`
            The parts that have been built by a robot.
        unbuilt_robot : list of :class:`compas_timber.beam.Blank`
            The parts that have not been built by a robot.

        """
        nodes = timber_assembly.graph.__data__["node"]
        buiding_plan_data_reference_list = [project_name, "building_plan", "data"]
        current_state_data = self.database.get_data_from_deep_reference(buiding_plan_data_reference_list)

        built_human = []
        unbuilt_human = []
        built_robot = []
        unbuilt_robot = []
        step_locations = []

        # Try to get the value for the last built index, if it doesn't exist make it null
        # TODO: This is a bit weird, but it will throw an error if I pass the last
        # TODO: built index to the BuildingPlan constructor
        if "LastBuiltIndex" in current_state_data:
            last_built_index = current_state_data["LastBuiltIndex"]
            current_state_data.pop("LastBuiltIndex")
        else:
            last_built_index = None
        if "PriorityTreeDictionary" in current_state_data:
            current_state_data.pop("PriorityTreeDictionary")

        if "PriorityTreeDictionary" in current_state_data:
            current_state_data.pop("PriorityTreeDictionary")

        building_plan = BuildingPlan.__from_data__(current_state_data)
        for step in building_plan.steps:
            step_data = step["data"]
            # Try to get the value for device_id, and if it exists remove it.
            if "device_id" in step_data:
                step_data.pop("device_id")
            step = Step.__from_data__(step["data"])
            step_locations.append(Frame.__from_data__(step.location))
            assembly_element_id = step.element_ids[0]
            # TODO: Tried to write like this, but find_by_key returns a NoneType object
            """
            part = timber_assembly.find_by_key(assembly_element_id)
            """
            part = nodes[assembly_element_id]["part"]
            if step.actor == "HUMAN":
                if step.is_built:
                    built_human.append(part.blank)
                else:
                    unbuilt_human.append(part.blank)
            else:
                if step.is_built:
                    built_robot.append(part.blank)
                else:
                    unbuilt_robot.append(part.blank)
        return last_built_index, step_locations, built_human, unbuilt_human, built_robot, unbuilt_robot

    def visualize_project_state(self, assembly, project_name):
        """
        Retrieves and visualizes data from the Firebase RealtimeDatabase under the specified project name.

        Parameters
        ----------
        assembly : :class:`compas.datastructure.Assembly`
            The assembly in which the project is based off of: Used for part visulization.
        project_name : str
            The name of the project under which the data is stored.

        Returns
        -------
        last_built_index : int
            The index of the last built part in the project.
        step_locations : list of :class:`compas.geometry.Frame`
            The locations of the building plan steps.
        built_human : list of :class:`compas.datastructures.Part`
            The parts that have been built by a human.
        unbuilt_human : list of :class:`compas.datastructures.Part`
            The parts that have not been built by a human.
        built_robot : list of :class:`compas.datastructures.Part`
            The parts that have been built by a robot.
        unbuilt_robot : list of :class:`compas.datastructures.Part`
            The parts that have not been built by a robot.

        """
        buiding_plan_data_reference_list = [project_name, "building_plan", "data"]
        current_state_data = self.database.get_data_from_deep_reference(buiding_plan_data_reference_list)
        nodes = assembly.graph.__data__["node"]

        built_human = []
        unbuilt_human = []
        built_robot = []
        unbuilt_robot = []
        step_locations = []

        # Try to get the value for the last built index, if it doesn't exist make it null
        # TODO: This is a bit weird, but it will throw an error if I pass the last built index to the BuildingPlan
        if "LastBuiltIndex" in current_state_data:
            last_built_index = current_state_data["LastBuiltIndex"]
            current_state_data.pop("LastBuiltIndex")
        else:
            last_built_index = None
        if "PriorityTreeDictionary" in current_state_data:
            current_state_data.pop("PriorityTreeDictionary")

        if "PriorityTreeDictionary" in current_state_data:
            current_state_data.pop("PriorityTreeDictionary")

        building_plan = BuildingPlan.__from_data__(current_state_data)
        for step in building_plan.steps:
            step_data = step["data"]
            # Try to get the value for device_id, and if it exists remove it.
            if "device_id" in step_data:
                step_data.pop("device_id")
            step = Step.__from_data__(step["data"])
            step_locations.append(Frame.__from_data__(step.location))
            assembly_element_id = step.element_ids[0]
            part = nodes[str(assembly_element_id)]["part"]

            if step.actor == "HUMAN":
                # TODO: I am not sure if this works in all scenarios of Part
                if step.is_built:
                    built_human.append(part)
                else:
                    unbuilt_human.append(part)
            elif step.actor == "ROBOT":
                if step.is_built:
                    built_robot.append(part)
                else:
                    unbuilt_robot.append(part)
            else:
                raise Exception("Part actor is Unknown!")
        return last_built_index, step_locations, built_human, unbuilt_human, built_robot, unbuilt_robot



import json
import os


class RealtimeDatabaseInterface(object):
    """
    The RealtimeDatabaseInterface class serves as the shared interface for RealtimeDatabase classes that
    operate in IronPython and Python 3.0.

    Methods within this class are designed to rely on shared interfaces that are implemented in child classes.
    """

    def construct_reference(self, parentname):
        raise NotImplementedError("Implemented on child classes")

    def construct_child_refrence(self, parentname, childname):
        raise NotImplementedError("Implemented on child classes")

    def construct_grandchild_refrence(self, parentname, childname, grandchildname):
        raise NotImplementedError("Implemented on child classes")

    def construct_reference_from_list(self, reference_list):
        raise NotImplementedError("Implemented on child classes")

    def upload_data_to_reference(self, data, database_reference):
        raise NotImplementedError("Implemented on child classes")

    def get_data_from_reference(self, database_reference):
        raise NotImplementedError("Implemented on child classes")

    def delete_data_from_reference(self, database_reference):
        raise NotImplementedError("Implemented on child classes")

    def stream_data_from_reference(self, callback, database_reference):
        raise NotImplementedError("Implemented on child classes")

    def upload_data(self, data, reference_name):
        """
        Uploads data to the Firebase Realtime Database under specified reference name.

        Parameters
        ----------
        data : Any
            The data to be uploaded, needs to be JSON serializable.
        reference_name : str
            The name of the reference under which the data will be stored.

        Returns
        -------
        None

        """
        database_reference = self.construct_reference(reference_name)
        self.upload_data_to_reference(data, database_reference)

    def upload_data_to_reference_as_child(self, data, reference_name, child_name):
        """
        Uploads data to the Firebase Realtime Database under specified reference name & child name.

        Parameters
        ----------
        data : Any
            The data to be uploaded, needs to be JSON serializable.
        reference_name : str
            The name of the reference under which the child should exist.
        child_name : str
            The name of the reference under which the data will be stored.

        Returns
        -------
        None

        """
        database_reference = self.construct_child_refrence(reference_name, child_name)
        self.upload_data_to_reference(data, database_reference)

    def upload_data_to_deep_reference(self, data, reference_list):
        """
        Uploads data to the Firebase Realtime Database under specified reference names in list order.

        Parameters
        ----------
        data : Any
            The data to be uploaded, needs to be JSON serializable.
        reference_list : list of str
            The names in sequence order in which the data should be nested for upload.

        Returns
        -------
        None

        """
        database_reference = self.construct_reference_from_list(reference_list)
        self.upload_data_to_reference(data, database_reference)

    def upload_data_from_file(self, path_local, refernce_name):
        """
        Uploads data to the Firebase Realtime Database under specified reference name from a file.

        Parameters
        ----------
        path_local : str
            The local path in which the data is stored as a json file.
        reference_name : str
            The name of the reference under which the data will be stored.

        Returns
        -------
        None

        """
        if not os.path.exists(path_local):
            raise Exception("path does not exist {}".format(path_local))
        with open(path_local) as config_file:
            data = json.load(config_file)
        database_reference = self.construct_reference(refernce_name)
        self.upload_data_to_reference(data, database_reference)

    def get_data(self, reference_name):
        """
        Retrieves data from the Firebase Realtime Database under the specified reference name.

        Parameters
        ----------
        reference_name : str
            The name of the reference under which the data is stored.

        Returns
        -------
        data : dict
            The retrieved data in dictionary format.

        """
        database_reference = self.construct_reference(reference_name)
        return self.get_data_from_reference(database_reference)

    def get_data_from_child_reference(self, reference_name, child_name):
        """
        Retreives data from the Firebase Realtime Database under specified reference name & child name.

        Parameters
        ----------
        reference_name : str
            The name of the reference under which the child exists.
        child_name : str
            The name of the reference under which the data is stored.

        Returns
        -------
        data : dict
            The retrieved data in dictionary format.

        """
        database_reference = self.construct_child_refrence(reference_name, child_name)
        return self.get_data_from_reference(database_reference)

    def get_data_from_deep_reference(self, reference_list):
        """
        Retreives data from the Firebase Realtime Database under specified reference names in list order.

        Parameters
        ----------
        data : Any
            The data to be uploaded, needs to be JSON serializable.
        reference_list : list of str
            The names in sequence order in which the is nested.

        Returns
        -------
        data : dict
            The retrieved data in dictionary format.
        """
        database_reference = self.construct_reference_from_list(reference_list)
        return self.get_data_from_reference(database_reference)

    def delete_data(self, reference_name):
        """
        Deletes data from the Firebase Realtime Database under specified reference name.

        Parameters
        ----------
        reference_name : str
            The name of the reference under which the child should exist.

        Returns
        -------
        None

        """
        database_reference = self.construct_reference(reference_name)
        self.delete_data_from_reference(database_reference)

    def delete_data_from_child_reference(self, reference_name, child_name):
        """
        Deletes data from the Firebase Realtime Database under specified reference name & child name.

        Parameters
        ----------
        reference_name : str
            The name of the reference under which the child should exist.
        child_name : str
            The name of the reference under which the data will be stored.

        Returns
        -------
        None

        """
        database_reference = self.construct_child_refrence(reference_name, child_name)
        self.delete_data_from_reference(database_reference)

    def delete_data_from_deep_reference(self, reference_list):
        """
        Deletes data from the Firebase Realtime Database under specified reference names in list order.

        Parameters
        ----------
        reference_list : list of str
            The names in sequence order in which the data should be nested for upload.

        Returns
        -------
        None

        """
        database_reference = self.construct_reference_from_list(reference_list)
        self.delete_data_from_reference(database_reference)



"""
Settings for Firebase.

COMPAS XR v1.0.0
"""

import json
import os

from ghpythonlib.componentbase import executingcomponent as component

from compas_xr.ghpython.firebase_config import FirebaseConfig


class FirebaseConfigComponent(component):
    def RunScript(self, filepath, filename, api_key, auth_domain, database_url, storage_bucket):
        if not (api_key and auth_domain and database_url and storage_bucket):
            self.Message = "You are missing some config information"
            raise Exception("Missing Config Info")

        config_path = None
        firebase_config = FirebaseConfig(api_key, auth_domain, database_url, storage_bucket)
        config = firebase_config.__data__()

        if config and filepath:
            if os.path.exists(filepath):
                if not filename:
                    filename = os.path.join(filepath, "firebase_config.json")
                if filename:
                    filename = os.path.join(filepath, filename)

                config = dict(config)
                with open(filename, "w") as f:
                    json.dump(config, f)
                config_path = filename
                self.Message = "Config Written"
            else:
                self.Message = "You filepath does not exist"
                raise Exception("Path does not exist {}".format(filepath))
        else:
            self.Message = "You are missing your filepath"
            raise Exception("Missing Filepath")

        return config_path



class FirebaseConfig(object):
    def __init__(self, api_key, auth_domain, database_url, storage_bucket):
        self.api_key = api_key
        self.auth_domain = auth_domain
        self.database_url = database_url
        self.storage_bucket = storage_bucket

    def ToString(self):
        return str(self)

    def __str__(self):
        return "FirebaseConfig, api_key={}, auth_domain={}, database_url={}, storage_bucket={}".format(self.api_key, self.auth_domain, self.database_url, self.storage_bucket)

    def __data__(self):
        return {
            "apiKey": self.api_key,
            "authDomain": self.auth_domain,
            "databaseURL": self.database_url,
            "storageBucket": self.storage_bucket,
        }



class AppSettings(object):
    def __init__(self, project_name, storage_folder=None, z_to_y_remap=None):
        self.project_name = project_name
        self.storage_folder = storage_folder or "None"
        self.z_to_y_remap = z_to_y_remap or False

    def ToString(self):
        return str(self)

    def __str__(self):
        return "AppSettings, project_name={}, storage_folder={}, z_to_y_remap={}".format(self.project_name, self.storage_folder, self.z_to_y_remap)

    def __data__(self):
        return {
            "project_name": self.project_name,
            "storage_folder": self.storage_folder,
            "z_to_y_remap": self.z_to_y_remap,
        }
